@@ -21,10 +21,7 @@ function w_normalize_filename($a) {
   if ($a === null) {
     return null;
   }
-  $a = str_replace('/', DIRECTORY_SEPARATOR, $a);
-  if (DIRECTORY_SEPARATOR == '\\') {
-    return strtolower($a);
-  }
+  $a = str_replace('\\', '/', $a);
   return $a;
 }
 
@@ -523,14 +520,15 @@ class WatchmanTestCase {
     sort($a);
     $b = w_normalize_file_list($b);
     sort($b);
-    $this->assertEqual($a, $b, $message);
+    $this->assertEqual($a, $b, $message . " " . json_encode($a) .
+                               " vs " .json_encode($b));
   }
 
   function assertFileList($root, array $files, $message = null) {
     $this->assertFileListUsingSince($root, null, $files, null, $message);
   }
 
-  private function secondLevelSort(array $objs) {
+   function secondLevelSort(array $objs) {
     $ret = array();
 
     foreach ($objs as $obj) {
