@@ -16,10 +16,6 @@ struct timeval {
   time_t tv_sec;
   time_t tv_usec;
 };
-struct timespec {
-  time_t tv_sec;
-  long long tv_nsec;
-};
 
 int gettimeofday(struct timeval *tv, void *ignored);
 void FILETIME_to_timespec(const FILETIME *ft, struct timespec *ts);
